# Copyright (c) 2022, Ora Lassila & So Many Aircraft
# All rights reserved.
#
# See LICENSE for licensing information
#
# This module implements some useful functionality for querying with RDFLib.
#
import logging
import rdflib
import rdflib.plugins.sparql.processor
from rdflib import Literal, URIRef
import string
import numbers

def identity(x):
    return x

# Make a new dictionary with values mapped using a callable
def mapDict(dictionary, mapper=identity):
    d = dict()
    for key, value in dictionary.items():
        d[key] = mapper(value)
    return d

# TEMPLATED QUERIES
#
# This mechanism can be used in lieu of RDFLib's "initBindings=" parameter for SPARQL queries,
# with the added benefit that replacements are not limited to SPARQL terms.
class Templated:

    @classmethod
    def query(cls, graph, template, **kwargs):
        q = cls.convert(template, kwargs) if kwargs else template
        logging.debug(q)
        return graph.query(q)

    @classmethod
    def update(cls, graph, template, **kwargs):
        q = cls.convert(template, kwargs) if kwargs else template
        logging.debug(q)
        graph.update(q)

    @classmethod
    def convert(cls, template, kwargs):
        return string.Template(template).substitute(**mapDict(kwargs, mapper=cls.forSPARQL))

    @classmethod
    def forSPARQL(cls, thing):
        if isinstance(thing, URIRef) or isinstance(thing, Literal):
            return thing.n3()
        elif isinstance(thing, str):
            return thing  # if thing[0] == '?' else cls.forSPARQL(Literal(thing))
        elif isinstance(thing, bool):
            return "true" if thing else "false"
        elif isinstance(thing, numbers.Number):
            return thing
        elif thing is None:
            return ""
        else:
            raise ValueError("Cannot make a SPARQL compatible value: %s", thing)

class TemplatedQueryMixin:  # abstract, can be mixed with Graph or Store

    def query(self, querystring, **kwargs):
        return Templated.query(super(), querystring, **kwargs)

    def update(self, querystring, **kwargs):
        Templated.update(super(), querystring, **kwargs)

class TemplateWrapper:
    def __init__(self, graph):
        self._graph = graph

    def query(self, querystring, **kwargs):
        return Templated.query(self._graph, querystring, **kwargs)

    def update(self, querystring, **kwargs):
        Templated.update(self._graph, querystring, **kwargs)


# Make a new Graph instance from triples (an iterable)
def graphFrom(triples, add_to=None, graph_class=rdflib.Graph, **kwargs):
    if add_to is None:
        add_to = graph_class(**kwargs)
    for triple in triples:
        add_to.add(triple)
    return add_to if len(add_to) > 0 else None

class Composable:
    def __init__(self, graph=None):
        if isinstance(graph, Composable):
            self._graph = graph._graph
        else:
            self._graph = graph or rdflib.Graph()

    @property
    def result(self):
        return self._graph

    def __len__(self):
        return len(self._graph)

    def __add__(self, other):
        if isinstance(other, Composable):
            other = other._graph
        elif not isinstance(other, rdflib.Graph):
            raise ValueError("Cannot be added to a graph: {}".format(other))
        return Composable(self._graph + other)

    def add(self, *triples):
        for triple in triples:
            self._graph.add(triple)
        return self

    def bind(self, prefix, namespace):
        self._graph.bind(prefix, namespace)
        return self

    @classmethod
    def parse(cls, *args, **kwargs):
        graph = rdflib.Graph()
        graph.parse(*args, **kwargs)
        return Composable(graph)

    def serialize(self, *args, **kwargs):
        self._graph.serialize(*args, **kwargs)

    def construct(self, template, **kwargs):
        # TODO: How do we confirm that `template` is a `CONSTRUCT` query?
        return self.__class__(graphFrom(Templated.query(self._graph, template, **kwargs)))

    def query(self, template, **kwargs):
        return Templated.query(self._graph, template, **kwargs)

    def update(self, template, **kwargs):
        Templated.update(self._graph, template, **kwargs)
        return self



# Copyright (c) 2022, Ora Lassila & So Many Aircraft
# All rights reserved.
#
# See LICENSE for licensing information
#
# This module implements the computation of the Concise Bounded Description.
#
from rdflib import Graph, ConjunctiveGraph, BNode
from rdfhelpers.templated import graphFrom

CBD_QUERY = '''
    SELECT ?s1 ?p1 ?o1 {
        {
            SELECT ?s1 ?p1 ?o1 {
                BIND($uri AS ?s1)
                ?s1 ?p1 ?o1
            }
        } UNION {
            SELECT ?s1 ?p1 ?o1 {
                BIND($uri AS ?s2)
                ?s2 ?p2 ?s1 . FILTER(ISBLANK(?s1))
                ?s1 ?p1 ?o1
            }
        } UNION {
            SELECT ?s1 ?p1 ?o1 {
                BIND($uri AS ?s3)
                ?s3 ?p3 ?s2 . FILTER(ISBLANK(?s2))
                ?s2 ?p2 ?s1 . FILTER(ISBLANK(?s1))
                ?s1 ?p1 ?o1
            }
        } UNION {
            SELECT ?s1 ?p1 ?o1 {
                BIND($uri AS ?s4)
                ?s4 ?p4 ?s3 . FILTER(ISBLANK(?s3))
                ?s3 ?p3 ?s2 . FILTER(ISBLANK(?s2))
                ?s2 ?p2 ?s1 . FILTER(ISBLANK(?s1))
                ?s1 ?p1 ?o1
            }
        }
    }

'''
CBD_QUERY_WITH_CONTEXT = '''
    SELECT ?s1 ?p1 ?o1 {
        GRAPH $context {
            {
                SELECT ?s1 ?p1 ?o1 {
                    BIND($uri AS ?s1)
                    ?s1 ?p1 ?o1
                }
            } UNION {
                SELECT ?s1 ?p1 ?o1 {
                    BIND($uri AS ?s2)
                    ?s2 ?p2 ?s1 . FILTER(ISBLANK(?s1))
                    ?s1 ?p1 ?o1
                }
            } UNION {
                SELECT ?s1 ?p1 ?o1 {
                    BIND($uri AS ?s3)
                    ?s3 ?p3 ?s2 . FILTER(ISBLANK(?s2))
                    ?s2 ?p2 ?s1 . FILTER(ISBLANK(?s1))
                    ?s1 ?p1 ?o1
                }
            } UNION {
                SELECT ?s1 ?p1 ?o1 {
                    BIND($uri AS ?s4)
                    ?s4 ?p4 ?s3 . FILTER(ISBLANK(?s3))
                    ?s3 ?p3 ?s2 . FILTER(ISBLANK(?s2))
                    ?s2 ?p2 ?s1 . FILTER(ISBLANK(?s1))
                    ?s1 ?p1 ?o1
                }
            }
        }
    }
'''

def cbd(source, resource,
        use_sparql=True, context=None, target=None, target_class=Graph, use_describe=False,
        **kwargs):
    if use_sparql:
        if context:
            triples = source.query(CBD_QUERY_WITH_CONTEXT, uri=resource, context=context)
        elif use_describe:
            triples = source.query("DESCRIBE $uri", uri=resource)
        else:
            triples = source.query(CBD_QUERY, uri=resource)
    else:
        # Best choice for a source graph with an in-memory store
        triples = _cbd_no_sparql(source, resource, context=context)
    return graphFrom(triples, add_to=target, graph_class=target_class, **kwargs)

def _cbd_no_sparql(source, resource, context=None):
    triples = list()
    subjects = list([resource])
    subjects_seen = list()
    if isinstance(source, ConjunctiveGraph):
        def get_triples(s): return source.quads((s, None, None, context))
    elif context is None:
        def get_triples(s): return source.triples((s, None, None))
    else:
        raise ValueError("Source graph must be context-aware")
    while subjects:
        subject = subjects.pop()
        if subject not in subjects_seen:
            subjects_seen.append(subject)
            for triple in get_triples(subject):
                triples.append(triple)
                leaf = triple[2]
                if isinstance(leaf, BNode):
                    subjects.append(leaf)
    return triples

REVERSE_CBD_QUERY = '''
    SELECT DISTINCT ?o1 ?p1 ?s1 {
        {
            SELECT ?o1 ?p1 ?s1 {
                BIND($uri AS ?s1)
                ?o1 ?p1 ?s1
                FILTER (!(?p1 = rdf:type && ?s1 in (rdfs:Resource, owl:Thing)))
            }
        } UNION {
            SELECT ?o1 ?p1 ?s1 {
                BIND($uri AS ?s2)
                ?s1 ?p2 ?s2 . FILTER(ISBLANK(?s1))
                ?o1 ?p1 ?s1
                FILTER (!(?p1 = rdf:type && ?s1 in (rdfs:Resource, owl:Thing)))
            }
        } UNION {
            SELECT ?o1 ?p1 ?s1 {
                BIND($uri AS ?s3)
                ?s2 ?p3 ?s3 . FILTER(ISBLANK(?s2))
                ?s1 ?p2 ?s2 . FILTER(ISBLANK(?s1))
                ?s1 ?p1 ?o1
                FILTER (!(?p1 = rdf:type && ?s1 in (rdfs:Resource, owl:Thing)))
            }
        } UNION {
            SELECT ?o1 ?p1 ?s1 {
                BIND($uri AS ?s4)
                ?s3 ?p4 ?s4 . FILTER(ISBLANK(?s3))
                ?s2 ?p3 ?s3 . FILTER(ISBLANK(?s2))
                ?s1 ?p2 ?s2 . FILTER(ISBLANK(?s1))
                ?s1 ?p1 ?o1
                FILTER (!(?p1 = rdf:type && ?s1 in (rdfs:Resource, owl:Thing)))
            }
        }
    }

'''
REVERSE_CBD_QUERY_WITH_CONTEXT = '''
    SELECT DISTINCT ?s1 ?p1 ?o1 {
        GRAPH $context {
            {
                SELECT ?o1 ?p1 ?s1 {
                    BIND($uri AS ?s1)
                    ?o1 ?p1 ?s1
                    FILTER (!(?p1 = rdf:type && ?s1 in (rdfs:Resource, owl:Thing)))
                }
            } UNION {
                SELECT ?o1 ?p1 ?s1 {
                    BIND($uri AS ?s2)
                    ?s1 ?p2 ?s2 . FILTER(ISBLANK(?s1))
                    ?o1 ?p1 ?s1
                    FILTER (!(?p1 = rdf:type && ?s1 in (rdfs:Resource, owl:Thing)))
                }
            } UNION {
                SELECT ?o1 ?p1 ?s1 {
                    BIND($uri AS ?s3)
                    ?s2 ?p3 ?s3 . FILTER(ISBLANK(?s2))
                    ?s1 ?p2 ?s2 . FILTER(ISBLANK(?s1))
                    ?s1 ?p1 ?o1
                    FILTER (!(?p1 = rdf:type && ?s1 in (rdfs:Resource, owl:Thing)))
                }
            } UNION {
                SELECT ?o1 ?p1 ?s1 {
                    BIND($uri AS ?s4)
                    ?s3 ?p4 ?s4 . FILTER(ISBLANK(?s3))
                    ?s2 ?p3 ?s3 . FILTER(ISBLANK(?s2))
                    ?s1 ?p2 ?s2 . FILTER(ISBLANK(?s1))
                    ?s1 ?p1 ?o1
                    FILTER (!(?p1 = rdf:type && ?s1 in (rdfs:Resource, owl:Thing)))
                }
            }
        }
    }
'''

def reverse_cbd(source, resource, context=None, target=None, target_class=Graph, **kwargs):
    if context:
        triples = source.query(REVERSE_CBD_QUERY_WITH_CONTEXT, uri=resource, context=context)
    else:
        triples = source.query(REVERSE_CBD_QUERY, uri=resource)
    return graphFrom(triples, add_to=target, graph_class=target_class, **kwargs)



# Copyright (c) 2022, Ora Lassila & So Many Aircraft
# All rights reserved.
#
# See LICENSE for licensing information
#
# This module implements some useful functionality for programming with RDFLib.
#

from rdflib import URIRef, BNode, Variable, Graph, RDF, RDFS, ConjunctiveGraph
from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore
from rdfhelpers.templated import Templated, graphFrom, TemplatedQueryMixin
from rdfhelpers.cbd import cbd
import re

# GLOSSARY
#
# This code uses certain terms or words in specific meanings:
# - "container" -- a composite object, an instance of any of the subclasses of rdfs:Container.
# - "context" -- something implemented as a named graph in the back-end.

# HELPFUL STUFF

def expandQName(prefix, local_name, ns_mgr):
    ns = ns_mgr.store.namespace(prefix)
    if ns is not None:
        return str(ns) + local_name
    else:
        raise KeyError("Namespace prefix {} is not bound".format(prefix))

def URI(u):
    return u if isinstance(u, URIRef) or u is None else URIRef(u)

def ntriples(graph, destination=None):
    result = graph.serialize(destination=destination, format="nt", encoding="utf-8")
    return result.decode("utf-8") if destination is None else None

# Helpful graph accessors

def getvalue(graph, node, predicate):
    return next(graph.objects(node, predicate), None)

def setvalue(graph, node, predicate, value):
    graph.remove((node, predicate, None))
    if value is not None:
        graph.add((node, predicate, value))

def addvalues(graph, node, predicates_and_values: dict):
    for predicate, value in predicates_and_values.items():
        graph.add((node, predicate, value))

def setvalues(graph, node, predicates_and_values: dict):
    for predicate, value in predicates_and_values.items():
        setvalue(graph, node, predicate, value)

def diff(graph1, graph2):
    return graph1 - graph2, graph2 - graph1

# CONTAINERS

LI_MATCH_PATTERN = re.compile(str(RDF) + "_([0-9]+)")
LI_CREATE_PATTERN = str(RDF) + "_{0}"

def isContainerItemPredicate(uri):
    match = LI_MATCH_PATTERN.match(uri)
    return int(match.group(1)) if match else None

def makeContainerItemPredicate(index):
    return LI_CREATE_PATTERN.format(index)

def getContainerStatements(graph, source, predicate):
    containers = list(graph.objects(URI(source), predicate))
    n = len(containers)
    if n == 1:
        return sorted([statement for statement in graph.triples((containers[0], None, None))
                       if isContainerItemPredicate(statement[1])],
                      key=lambda tr: tr[1])
    elif n == 0:
        return None
    else:
        raise ValueError("Expected only one value for {0}".format(predicate))

def getContainerItems(graph, node, predicate):
    statements = getContainerStatements(graph, node, predicate)
    return [statement[2] for statement in statements] if statements else None

def setContainerItems(graph, node, predicate, values, newtype=RDF.Seq):
    # Having to write code like this is a clear indication that triples are the wrong
    # abstraction for graphs, way too low level. Just sayin'.
    if values:
        statements = getContainerStatements(graph, node, predicate)
        if statements:
            container = statements[0][0]
            for statement in statements:
                graph.remove(statement)
        else:
            container = BNode()
            graph.add((node, predicate, container))
            graph.add((container, RDF.type, newtype))
        i = 1
        for value in values:
            graph.add((container, URIRef(makeContainerItemPredicate(i)), value))
            i += 1
    else:
        container = getvalue(graph, node, predicate)
        if container:
            graph.remove((node, predicate, container))
            graph.remove((container, None, None))

# FOCUSED GRAPH
#
# Instances of FocusedGraph have a specified "focus node".

class FocusedGraph(Graph):
    def __init__(self, focus=None, source=None, focus_class=None, **kwargs):
        super().__init__(**kwargs)
        if source:
            self.parse(source)
        self._focus = focus or self.findFocus(focus_class=focus_class, **kwargs)

    @property
    def focus(self):
        return self._focus

    def findFocus(self, focus_class=None, **kwargs):
        if focus_class:
            focus = next(self.triples((None, RDF.type, focus_class)), None)
            if focus:
                return focus[0]
        raise ValueError("No focus found")

    def getvalue(self, predicate):
        return getvalue(self, self._focus, predicate)

    def setvalue(self, predicate, value):
        setvalue(self, self._focus, predicate, value)

# CBD GRAPH

class CBDGraph(FocusedGraph):
    def __init__(self, focus, data, context=None, **kwargs):
        super().__init__(focus=focus, **kwargs)
        self.data = data
        self.context = context
        cbd(data, focus, target=self, context=context)

    def diff(self):
        context = self.context
        return diff(self, cbd(self, None, self.focus, context))

# SPARQL REPOSITORY

class SPARQLRepository(TemplatedQueryMixin, SPARQLUpdateStore):
    def __init__(self, query_endpoint=None, update_endpoint=None, **kwargs):
        super().__init__(query_endpoint=query_endpoint,
                         update_endpoint=update_endpoint or query_endpoint,
                         **kwargs)

    # Why do I get a complaint about this? SPARQLUpdateStore is not an abstract class...
    def triples_choices(self, _, context=None):
        pass



# Copyright (c) 2022, Ora Lassila & So Many Aircraft
# All rights reserved.
#
# See LICENSE for licensing information
#
# This module implements some useful functionality for programming with RDFLib.
#

import collections
import itertools
import re

from rdflib import URIRef, Literal
from rdfhelpers.templated import Templated

# LABEL CACHING
#
# Cache the values of rdfs:label (and sub-properties thereof) for faster access. Select
# either all vertices with a label (the default) or customize vertex selection.
class LabelCache:
    def __init__(self, db, label_query=None, selector=None, language="en", prepopulate=False):
        self.db = db
        self.label_query = label_query or self.LABELQUERY
        self.selector = selector or "?res ?p ?o"
        self.language = language
        self._labels = dict()
        self._search_index = collections.defaultdict(set)
        if prepopulate:
            self.repopulate()

    LABELQUERY = '''
        SELECT ?r ?label {
            $selector
            BIND ($resource AS ?r)
            ?label_prop rdfs:subPropertyOf* rdfs:label
            OPTIONAL { $resource skos:prefLabel ?pref_label }
            OPTIONAL { $resource skos:altLabel ?alt_label }
            OPTIONAL { $resource ?label_prop ?real_label }
            OPTIONAL { $resource dc:title ?other_label }
            BIND (COALESCE(?pref_label, ?alt_label, ?real_label, ?other_label) AS ?label)
            FILTER ((str(lang(?label)) = "") || langMatches(lang(?label), "$language"))
        }
        $limit
    '''

    ALLLABELSQUERY = '''
        SELECT DISTINCT ?label {
            $resource skos:prefLabel|skos:altLabel|dc:title|(rdfs:subPropertyOf*/rdfs:label) ?label
            BIND (IF(langMatches(lang(?label), "$language"), 0, 1) AS ?priority)
        }
        ORDER BY ?priority
    '''

    def getLabel(self, resource):
        if not isinstance(resource, URIRef):
            resource = URIRef(resource)
        label = self._labels.get(resource, None)
        if label:
            return label
        else:
            label = None
            for _, lb in Templated.query(self.db, self.label_query,
                                         resource=resource, selector=None, limit="LIMIT 1",
                                         language=self.language):
                label = lb
                self.updateSearchIndex(resource, label)
                break
            if label is None:
                label = self.makeQName(resource)
            if label is not None:
                self._labels[resource] = label
            return label or str(resource)

    def invalidateLabel(self, resource):
        self._labels[resource if isinstance(resource, URIRef) else URIRef(resource)] = None

    def repopulate(self):
        self._labels = {resource: label for resource, label
                        in Templated.query(self.db, self.label_query,
                                           resource="?res", selector=self.selector, limit=None,
                                           language=self.language)}
        self._search_index = collections.defaultdict(set)
        for resource, label in self._labels.items():
            self.updateSearchIndex(resource, label)

    def updateSearchIndex(self, resource, label):
        if label:
            for p in itertools.pairwise(label.upper()):
                self._search_index[p[0]+p[1]].add(resource)

    def findResources(self, substring):
        search_string = substring.upper()
        if len(substring) < 2:
            raise ValueError("Search string too short")
        if len(substring) == 2:
            return self._search_index[search_string]
        else:
            return [resource for resource in self.findResources(search_string[0:2])
                    if str(self.getLabel(resource)).upper().find(search_string) >= 0]

    def findResourcesForAutocomplete(self, substring):
        # this value can be "JSONified" (e.g., flask.json.jsonify)
        return [{"value": resource, "label": self.getLabel(resource)} for resource
                in self.findResources(substring)]

    def getAllLabels(self, resource, exclude_default_namespace=True):
        if not isinstance(resource, URIRef):
            resource = URIRef(resource)
        labels = [label for label, in Templated.query(self.db, self.ALLLABELSQUERY,
                                                      resource=resource, language=self.language)]
        qname = self.makeQName(resource)
        if qname == str(resource):
            return labels + [qname]
        elif exclude_default_namespace and qname.startswith(':'):
            return labels + [str(resource)]
        else:
            return labels + [qname, str(resource)]

    GENERIC_PREFIX_MATCHER = re.compile("ns[0-9]+")

    def makeQName(self, uri):
        try:
            prefix, _, name = self.db.namespace_manager.compute_qname(uri)
            if self.GENERIC_PREFIX_MATCHER.match(prefix):
                return str(uri)
            else:
                return "{}:{}".format(prefix, name)
        except ValueError:
            return str(uri)

# SKOS CONCEPT LABEL CACHING
#
# Similar to regular label caching, but the default selection of vertices includes only the
# concepts of a specified concept scheme. Also provides a mapping from labels to concepts.
class SKOSLabelCache(LabelCache):
    def __init__(self, db, scheme, selector=None):
        self._concepts = dict()
        selector = Templated.convert(selector or "?res skos:broader*/^skos:hasTopConcept $scheme",
                                     {"scheme": scheme})
        super().__init__(db, selector=selector)

    def invalidateLabel(self, resource):
        label = self.getLabel(resource)
        super().invalidateLabel(resource)
        self._concepts[label] = None

    def updateSearchIndex(self, resource, label):
        super().updateSearchIndex(resource, label)
        self._concepts[label] = resource

    def repopulate(self):
        self._concepts = dict()
        super().repopulate()

    def getConcept(self, label):
        return self._concepts.get(label if isinstance(label, Literal) else Literal(label), None)

    def listConcepts(self):
        return self._labels.keys()



# Copyright (c) 2022, Ora Lassila & So Many Aircraft
# All rights reserved.
#
# See LICENSE for licensing information
#
# This module implements some useful stuff when programming with RDFLib.

from rdfhelpers.rdfhelpers import expandQName, URI, getvalue, setvalue
from rdfhelpers.rdfhelpers import isContainerItemPredicate, makeContainerItemPredicate, diff
from rdfhelpers.cbd import cbd, reverse_cbd
from rdfhelpers.rdfhelpers import getContainerStatements, getContainerItems, setContainerItems
from rdfhelpers.rdfhelpers import SPARQLRepository, FocusedGraph, CBDGraph
from rdfhelpers.templated import graphFrom, identity, mapDict
from rdfhelpers.templated import Templated, Composable
from rdfhelpers.constructor import Constructor
from rdfhelpers.labels import LabelCache, SKOSLabelCache
from rdfhelpers.gsp import GraphStoreClient

__all__ = [
    'expandQName', 'URI', 'getvalue', 'setvalue',
    'isContainerItemPredicate', 'makeContainerItemPredicate', 'diff',
    'cbd', "reverse_cbd",
    'getContainerStatements', 'getContainerItems', 'setContainerItems',
    'SPARQLRepository', 'FocusedGraph', 'CBDGraph',
    'graphFrom', 'identity', 'mapDict',
    'Templated', 'Composable',
    'Constructor',
    'LabelCache', 'SKOSLabelCache',
    'GraphStoreClient'
]
